from importlib.resources import open_binary
import numpy as np
import struct

def get_mapping():
    '''
    Generates a mapping of Myanmar Unicode characters to their corresponding
    indices in the parameter array.

    Returns
    -------
    dict
        A mapping from Myanmar Unicode characters to indices.
    '''
    def get_chars(start, end):
        return [chr(char) for char in range(ord(start), ord(end) + 1)]

    chars = (get_chars('\u1000', '\u103F')
            + get_chars('\u104A', '\u109F')
            + get_chars('\uAA60', '\uAA7F')
            + get_chars('\uA9E0', '\uA9FF')
            + get_chars('\u2000', '\u200B'))

    return {char: i + 1 for i, char in enumerate(chars)}

def load_params():
    '''
    Loads parameters as a 2d array, which are log likelihood ratios of
    Unicode to Zawgyi.

    Returns
    -------
    numpy.ndarray
        Parameters as a 2d array.
    '''
    def read_char_array(f, size):
        return struct.unpack(f'{size}s', f.read(size))[0].decode('utf-8')

    def read_float(f):
        return struct.unpack('>f', f.read(4))[0]

    def read_int(f):
        return struct.unpack('>i', f.read(4))[0]

    def read_short(f):
        return struct.unpack('>h', f.read(2))[0]

    with open_binary('myanmartools.resources', 'zawgyiUnicodeModel.dat') as f:
        # check signature
        uzmodel_tag = read_char_array(f, 8)
        if uzmodel_tag != 'UZMODEL ':
            raise IOError('incorrect uzmodel_tag')
        uzmodel_version = read_int(f)
        if uzmodel_version == 1:
            ssv = 0
        elif uzmodel_version == 2:
            ssv = read_int(f)
        else:
            raise IOError('incorrect uzmodel_version')
        bmarkov_tag = read_char_array(f, 8)
        if bmarkov_tag != 'BMARKOV ':
            raise IOError('incorrect bmarkov_tag')
        bmarkov_version = read_int(f)
        if bmarkov_version != 0:
            raise IOError('incorrect bmarkov_version')

        # read params
        size = read_short(f)
        params = np.empty((size, size))
        for row in range(size):
            count = read_short(f)
            if count != 0:
                params[row] = read_float(f)
                for i in range(count):
                    col = read_short(f)
                    params[row, col] = read_float(f)
            else:
                params[row] = 0

        return params



import numpy as np
from ._params import get_mapping, load_params

class ZawgyiDetector:
    '''
    An estimator that predicts Zawgyi using two Markov chains, one
    for Unicode text and the other for Zawgyi text.

    Attributes
    ----------
    mapping : dict
        A mapping of Myanmar Unicode characters to the corresponding indices
        in the parameter array.
    params : numpy.ndarray
        A parameter array containing log likelihood ratios of
        Unicode to Zawgyi.
    '''
    def __init__(self):
        self.mapping = get_mapping()
        self.params = load_params()

    def get_zawgyi_probability(self, string):
        '''
        Computes Zawgyi probability.

        Parameters
        ----------
        string : str
            String to predict Zawgyi on.

        Returns
        -------
        float
            Zawgyi probability between 0 and 1, or negative infinity
            if there is no Myanmar Unicode character.
        '''
        indices = [self.mapping.get(char, 0) for char in string]
        # include starting and ending state probabilities
        previous = np.array([0] + indices)
        current = np.array(indices + [0])
        # ignore 0-to-0 transitions
        mask = np.logical_or(previous != 0, current != 0)
        # return negative inifinity if there are only 0-to-0 transitions,
        # which happens when there is no Myanmar Unicode character
        if not mask.any():
            return -np.inf
        # Pz/(Pu+Pz) = exp(logPz)/(exp(logPu)+exp(logPz))
        #            = 1/(1+exp(logPu-logPz))
        return 1.0 / (1.0 + 
            np.exp(self.params[previous[mask], current[mask]].sum()))



'''
Myanmar Tools
=============

Myanmar Tools implements tools for processing font encodings used in Myanmar.
It currently supports Zawgyi detection.

To detect Zawgyi, create an instance of ZawgyiDetector, and call
``get_zawgyi_probability`` with a string::

    from myanmartools import ZawgyiDetector

    detector = ZawgyiDetector()
    score = detector.get_zawgyi_probability('မ္း')
    # score is now 0.999772 (very likely Zawgyi)

For Zawgyi-to-Unicode conversion, you can use the ICU library. Install it
using ``pip install PyICU``.

To convert Zawgyi to Unicode, create an instance of ICU Transliterator with
the transform ID "Zawgyi-my", and call :code:`transiliterate` with a string::

    from icu import Transliterator

    converter = Transliterator.createInstance('Zawgyi-my')
    output = converter.transliterate('မ္း')
    # output is now 'မ်း'
'''
from .zawgyi_detector import ZawgyiDetector



import csv, unittest
from importlib.resources import open_text
from myanmartools import ZawgyiDetector

class TestZawgyiDetector(unittest.TestCase):
    '''
    Tests the ZawgyiDetector class.
    '''
    def setUp(self):
        '''
        Creates a ZawgyiDetector object for testing.
        '''
        self.detector = ZawgyiDetector()

    def test_strong(self):
        '''
        Tests a Unicode string which should produce a very low score
        and a Zawgyi string which should produce a very high score.
        '''
        strong_unicode = 'အပြည်ပြည်ဆိုင်ရာ လူ့အခွင့်အရေး ကြေညာစာတမ်း'
        strong_zawgyi = 'အျပည္ျပည္ဆိုင္ရာ လူ႔အခြင့္အေရး ေၾကညာစာတမ္း'
        self.assertLess(
            self.detector.get_zawgyi_probability(strong_unicode),
            0.001
        )
        self.assertGreater(
            self.detector.get_zawgyi_probability(strong_zawgyi),
            0.999
        )

    def test_compatibility(self):
        '''
        Compares results with those obtained from Java version of the
        detector. They should be the same, within float tolerances.
        '''
        with open_text('myanmartools.resources', 'compatibility.tsv') as f:
            for row in csv.reader(f, delimiter='\t'):
                self.assertAlmostEqual(
                    self.detector.get_zawgyi_probability(row[1]),
                    float(row[0])
                )

if __name__ == '__main__':
    unittest.main()



import setuptools

with open('README.rst') as f:
    long_description = f.read()

setuptools.setup(
    name='myanmartools',
    version='0.1',
    url='https://github.com/google/myanmar-tools',
    author='William (Wai Yan) Zhu',
    author_email='williamzhu345@gmail.com',
    classifiers=[
	'Development Status :: 4 - Beta',
	'Intended Audience :: Developers',
	'Intended Audience :: Science/Research',
	'License :: OSI Approved :: Apache Software License',
	'Operating System :: OS Independent',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: 3.8',
	'Topic :: Text Processing'
    ],
    license='Apache License, Version 2.0',
    description='Tools for processing font encodings used in Myanmar',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    keywords=['burmese', 'encoding', 'myanmar', 'nlp', 'unicode', 'zawgyi'],
    platforms=['any'],
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    package_data={'myanmartools': ['resources/*']},
    install_requires=['numpy>=1.18'],
    python_requires='>=3.7'
)
